
import time
import copy
import matplotlib
import progressbar
import numpy as np
from loguru import logger

try:
    matplotlib.use('TKAgg')
    from matplotlib import pyplot as plt
except:
    pass


class Profiling(object):
    """
    The class Profiling is designed for perfomance test.

    Example:
        def fibonacci(i):
            if i > 0 and isinstance(i, int):
                num_list = [0, 1]
                if i < 2:
                    return num_list[i]
                elif i >= 2:
                    return (fibonacci(i - 2) + fibonacci(i - 1))
            else:
                return -1

        p = Profiling(action=fibonacci)

        @p.inputs_generator(name="g1", TYPE="A", MIN=1000, MAX=10000, STEP=1000)
        def generator1(index):
            return 11

        @p.inputs_generator(name="g2", TYPE="B", PER=1, GROUP=30)
        def generator2(index):
            return index

        p.run(generator_name="g1", times=5)
        p.run(generator_name="g2", times=5)
        p.plot("g2", 'g2.png')
        p.plot("g1", 'g1.png')

    :param action: the function that will be profiled
    """

    def __init__(self, action=None):
        self.action = action or (lambda x: x)
        self.generator = dict()
        self.pbar = progressbar.ProgressBar(
            max_value=progressbar.UnknownLength)

    def inputs_generator(self, name, TYPE, MIN=10, MAX=100, STEP=10, PER=1, GROUP=1):
        '''
        Decorator for function to generate the inputs

        Example:
            p = Profiling(action=fibonacci)

            @p.inputs_generator(name="g1", TYPE="A", MIN=1000, MAX=10000, STEP=1000)
            def generator1(index):
                return 11


            @p.inputs_generator(name="g2", TYPE="B", PER=1, GROUP=30)
            def generator2(index):
                return index

        :param name: the name of generator
        :param TYPE: the type of generator, and the value could only be ``A`` or ``B``.
            If type is ``A``, ``MIN``, ``MAX`` and ``STEP`` are required. And if type is ``B``, ``PER`` and ``GROUP`` are required
        :param MIN: the minimal call times
        :param MAX: the maximal call times
        :param STEP: the increment times of every step
        :param PER: the call time of every group
        :param GROUP: the mumber of groups

        :returns: Profiling.generator
        '''
        def wrapper(func):
            self.generator[name] = {
                "generator": func,
                "type": TYPE,
                "result": []
            }

            if TYPE == "A":
                self.generator[name]["min"] = MIN if MIN >= 0 else 0
                self.generator[name]["max"] = MAX if MIN <= MAX else MIN
                self.generator[name]["step"] = STEP if STEP <= MAX - \
                    MIN else MAX - MIN
            elif TYPE == "B":
                self.generator[name]["per"] = PER if PER >= 0 else 0
                self.generator[name]["group"] = GROUP if GROUP >= 1 else 1

        return wrapper

    def run(self, generator_name, times=1):
        """
        Schedule the profiling program to run

        :param generator_name: the name of generator 
        :param times: the running times for report
        """
        params = self.generator[generator_name]

        for k in range(times):
            result = {
                "x": [],
                "y": []
            }
            current = 0
            if params["type"] == "A":

                total = sum(
                    list(range(params["min"], params["max"], params["step"])))
                self.pbar = progressbar.ProgressBar(
                    max_value=total)
                logger.info(
                    "Selected generator[name:{}], times: {},  Total call times: {}", generator_name, k+1, total)
                for i in range(params["min"], params["max"], params["step"]):
                    start = time.time()
                    for j in range(i):
                        self.action(params["generator"](index=i))
                        current += 1
                        self.pbar.update(current)
                    time_cost = time.time() - start
                    result["x"].append(i)
                    result["y"].append(time_cost)
                self.generator[generator_name]["result"].append(
                    copy.deepcopy(result))

            elif params["type"] == "B":
                total = params["group"] * params["per"]
                logger.info(
                    "Selected generator[name:{}], times: {},  Total call times: {}", generator_name, k+1, total)
                self.pbar = progressbar.ProgressBar(
                    max_value=total)
                for i in range(params["group"]):
                    start = time.time()
                    for j in range(params["per"]):
                        self.action(params["generator"](index=i))
                        current += 1
                        self.pbar.update(current)
                    time_cost = time.time() - start
                    result["x"].append(i)
                    result["y"].append(time_cost)
                self.generator[generator_name]["result"].append(
                    copy.deepcopy(result))

    def plot(self, name, outfile=None):
        """
        Plot the profiling figure

        :param name: the name of generator 
        :param outfile: the filepath of output figure
        """
        plt.cla()
        plt.title(name)
        plt.ylabel("time cost(s)")
        for data in self.generator[name]["result"]:
            plt.plot(data["x"], data["y"])
        if outfile:
            plt.savefig(outfile)
        else:
            plt.show()



import math
from dataclasses import dataclass

def task_divide(data, granularity):
    l = len(data)
    if l < granularity:
        return [data]
    n = int(len(data) / granularity)
    return [data[i::n] for i in range(n)]


@dataclass(init=True, repr=True, frozen=True)
class Task:
    task_id: str
    branch_id: str
    data_type: str
    position: int
    properties: dict
    end_condition: dict





from loguru import logger
import asyncio
from functools import partial
import signal
import traceback
from ..utils.async_pool import AsyncPool
import multiprocessing
import concurrent.futures

class TimeoutError(Exception):
    pass

class KafkaError(Exception):
    pass

class EnptyInputError(Exception):
    pass


class EmptyOutputError(Exception):
    pass


def default_process(inputs, config):
    return inputs


def time_limit_function(func, time_limit: int, *args, raise_error = False,  **kwargs):
    """
    Execute a function with time limit
    :param func: (callable object) The function to be executed
    :param time_limit: (int) Time limit
    :param args:  the params of the function
    :param kwargs: the params of the function
    :return:
    """

    def callback(signum, frame):
        if raise_error:
            raise TimeoutError
        else:
            return None

    signal.signal(signal.SIGALRM, callback)
    signal.alarm(time_limit)
    result = func(*args, **kwargs)
    signal.alarm(0)
    return result



def process_data(data, *args, func=default_process, timeout=1, mode="loop", pool_size=None, **kwargs):
    """
    Process data with different mode
    :param data: input data, iterable object
    :param func: process function
    :param timeout: process timeout for each input
    :param mode: process mode , one of [total, loop, async, thread, process]
    :param pool_size: pool size for mode async, thread, and process
    :param args: other params for func
    :param kwargs: other params for func
    :return:
    """
    def process_collect_result(result):
        if result is not None:
            results.append(result)

    def process_collect_error(error):
        logger.warning(error.__repr__())
        errors.append(error.__repr__())

    if mode not in ["total", "loop", "async", "thread", "process"]:
        raise ValueError("Unsupported process mod: %s, should be one of [total, loop, async, thread, process]" % mode)
    if mode in ["async", "thread", "process"] and pool_size is None:
        DEFAULT_POOL_SIZE = {"async": 100, "thread": 100, "process": 10}
        logger.warning("Unspecified pool size, use default for %s: %d", DEFAULT_POOL_SIZE[mode])
        pool_size = DEFAULT_POOL_SIZE[mode]

    results = []
    errors = []
    # switch mode
    if mode == "total":
        results = time_limit_function(func, timeout * len(data), data, *args, **kwargs)
    elif mode == "loop":
        for item in data:
            try:
                r = time_limit_function(func, timeout, item, *args, **kwargs)
                if r is not None:
                    results.append(r)
            except Exception as err:
                logger.exception(err)
                errors.append(traceback.format_exc())
    elif mode == "thread":
        pool = concurrent.futures.ThreadPoolExecutor(max_workers=pool_size)
        tasks = [ pool.submit(func, item, *args, **kwargs) for item in data]
        for task in tasks:
            try:
                r = task.result(timeout=timeout)
                if r is not None:
                    results.append(r)
            except:
                traceback.print_exc()
                errors.append(traceback.format_exc())
    elif mode == "process":
        pool = multiprocessing.Pool(processes=pool_size)
        result = pool.map_async(partial(time_limit_function, **kwargs), [(func, timeout, item)+args for item in data],
                                callback=process_collect_result, error_callback=process_collect_error)
        result.wait()
    elif mode == "async":
        pool = AsyncPool(pool_size)
        pool.map_async(func, [(item,)+args for item in data], timeout,
                       callback=process_collect_result, error_callback=process_collect_error)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(pool.wait())

    return results, errors




# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.
print(find_packages("autumn"))
print(find_packages(""))
setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install sampleproject
    #
    # And where it will live on PyPI: https://pypi.org/project/sampleproject/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    name='e-autumn',  # Required

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    #
    # For a discussion on single-sourcing the version across setup.py and the
    # project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.0.30',  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='A executor framework',  # Optional

    # This is an optional longer description of your project that represents
    # the body of text which users will see when they visit PyPI.
    #
    # Often, this is the same as your README, so you can just read it in from
    # that file directly (as we have already done above)
    #
    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional

    # Denotes that our long_description is in Markdown; valid values are
    # text/plain, text/x-rst, and text/markdown
    #
    # Optional if long_description is written in reStructuredText (rst) but
    # required for plain-text or Markdown; if unspecified, "applications should
    # attempt to render [the long_description] as text/x-rst; charset=UTF-8 and
    # fall back to text/plain if it is not valid rst" (see link below)
    #
    # This field corresponds to the "Description-Content-Type" metadata field:

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    

    # This should be your name or the name of the organization which owns the
    # project.
    author='Druggle',  # Optional

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email='iamdrug@qq.com',  # Optional
    url = 'https://coding.net/u/Alecyrus/p/Autumn',
    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        'Programming Language :: Python :: 3.7'
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='executor',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=["autumn", "autumn.lib", "autumn.utils"],  # Required

    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. If you
    # do not support Python 2, you can simplify this to '>=3.5' or similar, see
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires='>=3.7',

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['tenacity>=5.0.4','colorama>=0.3.7', 'Flask>=1.0.3', 'termcolor>=1.1.0', 
                      'fire>=0.1.3', 'emoji>=0.5.2', 'pyfiglet>=0.8', 'requests>=2.22.0', 'PyYAML>=3.11', 'loguru>=0.2.5',
                      'pygments>=2.4.2', 'confluent_kafka>=0.11.6', 'async_timeout>=3.0.1']
)



import socket
from loguru import logger


SO_BINDTODEVICE = 25

# iface参数指Linux的网卡接口，如(eth0,wlan0)，这个参数只支持Linux并且需要root权限


def get_free_port(iface=None):
    """
    Get a free port

    :param iface: the interface of network devices: eth0
    :returns: port
    """
    s = socket.socket()
    if iface:
        s.setsockopt(socket.SOL_SOCKET, SO_BINDTODEVICE, bytes(iface, 'utf8'))
    s.bind(('', 0))
    port = s.getsockname()[1]
    s.close()
    logger.debug("Get a free port [{}]", port)
    return port
